import fastapi
from fastapi import Depends, HTTPException, Response
from auth import auth
import logging
import conf
import json
from starlette_context import context
from starlette.requests import Request
from bundles import bundle
import correlation
from pydantic import BaseModel

from starlette.status import (
    HTTP_201_CREATED,
    HTTP_202_ACCEPTED,
    HTTP_400_BAD_REQUEST,
    HTTP_405_METHOD_NOT_ALLOWED,
    HTTP_424_FAILED_DEPENDENCY,
)

from audit_logger import write_audit_log
from string_logger import StringLogger


class Detail(BaseModel):
    detail: str


router = fastapi.APIRouter()

logger = logging.getLogger(__name__)

description = f"""
## Experimental bootstrap API for creating and updating bundle to default.

This should be used when adding a partition to OSDU or when wanting to reset to a known good state.
    
### Without force:

- This method is only allowed if the partition doesn't already have a bundle.
- If the bundle already exists it will return 405 METHOD_NOT_ALLOWED.
- Policy Service can be configured to ignore force.

### May return:

- `HTTP_202_ACCEPTED` - updated
- `HTTP_201_CREATED` - created
- `HTTP_405_METHOD_NOT_ALLOWED` - not allowed
- `HTTP_424_FAILED_DEPENDENCY` - bundle server caused failure

### Permission required to use this API on this system (configurable):

    {conf.OPS_PERMISSION}
    """


@router.post(
    "/bootstrap",
    responses={
        HTTP_201_CREATED: {"model": Detail},
        HTTP_202_ACCEPTED: {"model": Detail},
        HTTP_400_BAD_REQUEST: {"model": Detail},
        HTTP_405_METHOD_NOT_ALLOWED: {"model": Detail},
    },
    description=description,
)
def bootstrap(
    response: Response,
    request: Request,
    auth_data: auth.Auth = Depends(auth.require_authorized_ops),
    force: bool = False,
):
    """
    Experimental bootstrap API for creating and updating bundle to default.
    """

    logging.setLogRecordFactory(
        correlation.set_correlation_id(context["correlation_id"])
    )
    logger = logging.getLogger(__name__)
    audit_log = StringLogger("AUDIT")

    response.headers["X-Correlation-ID"] = context["correlation_id"]

    # If uvicorn is running with proxy headers enabled (--proxy-headers)
    x_forwarded_for = request.headers.get("x-forwarded-for")

    # x-request-id
    x_request_id = request.headers.get("x-request-id")

    # Zipkin
    # B3 Propagation is a specification for the header "b3" and those that start with "x-b3-".
    # These headers are used for trace context propagation across service boundaries.
    x_b3_traceid = request.headers.get("x-b3-traceid")
    x_b3_spanid = request.headers.get("x-b3-spanid")
    x_b3_parentspanid = request.headers.get("x-b3-parentspanid")
    audit_details = f"correlation-id: {context['correlation_id']} user: {auth_data.user_id}  xuserid: {auth_data.x_user_id} requestid: {x_request_id} traceid: {x_b3_traceid} spanid: {x_b3_spanid} parentspanid: {x_b3_parentspanid} forward: {x_forwarded_for}"

    if bundle.get_bundle(data_partition=auth_data.data_partition_id):
        logger.debug("bootstrap bundle already exists")
        if force:
            if not conf.ALLOW_FORCE_BOOTSTRAP:
                raise HTTPException(
                    status_code=HTTP_405_METHOD_NOT_ALLOWED,
                    detail=f"'{auth_data.data_partition_id}' already bootstrapped and forced updates not allowed.",
                )
            status = bootstrap_bundle(data_partition=auth_data.data_partition_id)
            if status:
                response.status_code = HTTP_202_ACCEPTED
                detail = f"bootstrap bundle for partition '{auth_data.data_partition_id}' updated to default with force"
                audit_log_msg = f"{detail}, {audit_details}"
                audit_log.logger.info(audit_log_msg)
                logger.info(f"AUDIT {audit_log_msg}")
                write_audit_log(
                    data_partition=auth_data.data_partition_id,
                    log_message=str(audit_log),
                )
                return {"detail": detail}
            else:
                detail = f"failed-bootstrap bundle for partition '{auth_data.data_partition_id}', force update failed {status}!"
                audit_log_msg = f"{detail}, {audit_details}"
                audit_log.logger.error(audit_log_msg)
                logger.error(audit_log_msg)
                write_audit_log(
                    data_partition=auth_data.data_partition_id,
                    log_message=str(audit_log),
                )
                raise HTTPException(
                    status_code=HTTP_424_FAILED_DEPENDENCY,
                    detail=detail,
                )
        else:
            detail = f"failed-bootstrap bundle for partition '{auth_data.data_partition_id}', update failed {HTTP_405_METHOD_NOT_ALLOWED} already bootstrapped!"
            audit_log_msg = f"{detail}, {audit_details}"
            audit_log.logger.error(audit_log_msg)
            logger.error(f"AUDIT {audit_log_msg}")
            write_audit_log(
                data_partition=auth_data.data_partition_id, log_message=str(audit_log)
            )
            raise HTTPException(
                status_code=HTTP_405_METHOD_NOT_ALLOWED,
                detail=detail,
            )
    else:
        status = bootstrap_bundle(data_partition=auth_data.data_partition_id)
        if status:
            response.status_code = HTTP_201_CREATED
            detail = f"bootstrap bundle for partition '{auth_data.data_partition_id}' created"
            audit_log_msg = f"{detail}, {audit_details}"
            logger.info(f"AUDIT {audit_log_msg}")
            audit_log.logger.info(audit_log_msg)
            write_audit_log(
                data_partition=auth_data.data_partition_id, log_message=str(audit_log)
            )
            return {"detail": detail}
        else:
            detail = f"failed-bootstrap bundle for partition '{auth_data.data_partition_id}', create failed {status}!"
            audit_log_msg = f"{detail}, {audit_details}"
            audit_log.logger.error(audit_log_msg)
            logger.error(f"AUDIT {audit_log_msg}")
            write_audit_log(
                data_partition=auth_data.data_partition_id, log_message=str(audit_log)
            )
            raise HTTPException(status_code=HTTP_424_FAILED_DEPENDENCY, detail=detail)


def bootstrap_bundle(data_partition):
    # default dataauthz rego for a partition
    dataauthz = """\
package osdu.partition["{}"].dataauthz

import data.osdu.instance.dataauthz as centralauthz

records := centralauthz.records""".format(data_partition)

    # default search rego for a partition
    search = """\
package osdu.partition["{}"].search

default allow = false

allow = true {{
    input.operation == "view"
    # At least one user group needs to be in acl viewers
    input.record.acl.viewers[_]==input.groups[_]
}}

allow = true {{
    input.operation == ["view", "create", "update", "delete", "purge"][_]
    # At least one user group needs to be in acl owners
    input.record.acl.owners[_]==input.groups[_]
}}""".format(data_partition)

    # manifest is json
    manifest = {"roots": [f"osdu/partition/{data_partition}"]}

    #    manifest = """\
    # {{
    #    "roots": ["osdu/partition/{}"]
    # }}""".format(data_partition)

    default_policy_list = {
        # ".manifest": manifest.encode('ascii'),
        ".manifest": json.dumps(manifest, indent=0).encode("ascii"),
        "dataauthz.rego": dataauthz.encode("ascii"),
        "search.rego": search.encode("ascii"),
    }

    return bundle.put_default_bundle(
        data_partition=data_partition, policy_list=default_policy_list
    )
