#  Copyright 2022 Google LLC
#  Copyright 2022 EPAM Systems
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

import abc
import importlib
import logging
import os
import sys
from typing import Tuple

from osdu_api.providers.types import FileLikeObject

logger = logging.getLogger(__name__)
#logger = correlation.DefaultExtrasAdapter(_logger, {"correlation_id": "None"})

class BundleStorageClient(abc.ABC):
    """Base interface for bundle storage clients"""

    @abc.abstractmethod
    def download_file(self, filename: str, file: FileLikeObject) -> Tuple[FileLikeObject, str]:
        """Download file by name"""
        logger.debug(f"download_file {filename}")
        pass

    @abc.abstractmethod
    def upload_file(self, name: str, file: FileLikeObject) -> bool:
        """Upload file by name"""
        logger.debug(f"upload_file {name}")
        pass

def _import_provider_specific_module(provider: str) -> str:
    """Import provider specific module"""
    module_name = f"bundles.providers.{provider.lower()}"
    module = importlib.import_module(module_name)
    return module

def get_storage() -> BundleStorageClient:
    try:
        os.environ["CLOUD_PROVIDER"]
    except KeyError:
        logger.critical("Error: No CLOUD_PROVIDER ENV VAR SET")
        sys.exit(1)

    cloud_env = os.environ.get("CLOUD_PROVIDER")
    try:
        provider_module = _import_provider_specific_module(cloud_env)
    except ModuleNotFoundError as exc:
        logger.critical(f"Error occurred while importing module for {cloud_env}")
        logger.critical(f"Exception: {exc}")

    logger.info(f"get_storage provider for cloud_env {cloud_env}")
    return provider_module.storage_client()
