# Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import argparse
import boto3
import logging
import os
import tarfile
from jinja2 import Environment, FileSystemLoader, select_autoescape

LOG_FORMAT = "[%(filename)s:%(lineno)d|%(levelname)s] %(message)s"

# Constants
DATAAUTHZ_TEMPLATE_NAME = 'dataauthz_template.rego'
DATAAUTHZ_FILENAME = 'dataauthz.rego'
SEARCH_TEMPLATE_NAME = 'search_template.rego'
SEARCH_FILENAME = 'search.rego'
MANIFEST_TEMPLATE_NAME = 'manifest_template.manifest'
MANIFEST_FILENAME = '.manifest'

DEFAULT_POLICIES = {'legal.rego', 'entitlements.rego', 'search.rego',
                    'dataauthz.rego', '.manifest'}

# Template paths
DEFAULT_POLICIES_PATH = os.path.dirname(os.path.dirname(os.path.dirname(
    os.path.realpath(__file__)))) + '/default-policies'
DEFAULT_POLICY_TEMPLATES_PATH = os.path.dirname(
    os.path.realpath(__file__))+'/templates'
DEFAULT_POLICY_CONFIG_MAP_TEMPLATE_PATH = os.path.dirname(
    os.path.realpath(__file__))+'/templates'

# Parse command line arguments, both required and optional (some with defaults to fall back on)
parser = argparse.ArgumentParser(
    description="Initializes the policy bootstrapping utility")

parser.add_argument("region",
                    type=str,
                    help="The AWS region of the deployment")

parser.add_argument("policy_bucket",
                    type=str,
                    help="The S3 policy bucket name")

parser.add_argument("--partitions",
                    nargs="+",
                    type=str,
                    default=[],
                    help="A list of partitions with policies")

parser.add_argument("--profile",
                    type=str,
                    help="The AWS CLI profile to use")

parser.add_argument("--debug",
                    type=bool,
                    default=False,
                    help="Sets logger level to debug")

# Parse provided command line arguments
args = parser.parse_args()


class AwsPolicyBootstrap:
    """Utility class that provides policy bootstrapping functions."""

    def __init__(
        self,
        region: str,
        profile: str,
        policy_bucket: str,
        partitions: list,
        debug=False
    ):
        if debug:
            logging_level = logging.DEBUG
        else:
            logging_level = logging.INFO
        logging.basicConfig(format=LOG_FORMAT, level=logging_level)
        self.log = logging.getLogger("POLICY")

        self.region = region
        self.policy_bucket = policy_bucket
        self.partitions = partitions

        # Boto S3 client setup - S3 client requires resource level permissions
        boto3.setup_default_session(
            profile_name=profile, region_name=region)
        self.s3 = boto3.resource('s3')

    def _upload_to_policy_bucket(self, file: str) -> None:
        """Uploads the policy bundle from a local file to the corresponding S3 bucket.

        Requires a file name to upload.
        """

        self.log.info(
            f'RUNNING: Uploading file: {file} to Policy S3 bucket: {self.policy_bucket}...')

        try:
            self.s3.Object(
                self.policy_bucket, file).put(Body=open(file, 'rb'))
            self.log.info(
                f"SUCCESS: Upload of the policy bundle {file} to policy bucket was successful.")
        except RuntimeError as e:
            self.log.info(
                f"FAILED: There was an exception while attempting to upload the policy bundle {file} to the policy bucket.")
            raise e

    def upsert_instance_policy_bundle(self, path: str) -> None:
        """Creates a bundle of instance policies given the default bundle path."""

        file_list = []

        tar_name = f'bundle.tar.gz'
        with tarfile.open(tar_name, "w:gz") as tar_handle:
            for root, dirs, files in os.walk(path):
                for file in files:
                    if file in DEFAULT_POLICIES:
                        tar_handle.add(os.path.join(root, file), arcname=file)
                        file_list.append(file)
        self.log.info(
            f"Adding files: {', '.join(file_list)} to bundle: {tar_name}")
        self._upload_to_policy_bucket(tar_name)
        os.remove(tar_name)

    def upsert_partition_policy_bundles(self, path: str) -> None:
        """Creates a bundle of partition policies given the default template path."""

        env = Environment(
            loader=FileSystemLoader(path),
            autoescape=select_autoescape()
        )

        for dp_id in self.partitions:
            tar_name = f'bundle-{dp_id}.tar.gz'
            dataauthz_template = env.get_template(DATAAUTHZ_TEMPLATE_NAME)
            search_template = env.get_template(SEARCH_TEMPLATE_NAME)
            manifest_template = env.get_template(MANIFEST_TEMPLATE_NAME)
            dataauthz_render = dataauthz_template.render(dp_id=dp_id)
            search_render = search_template.render(dp_id=dp_id)
            manifest_render = manifest_template.render(dp_id=dp_id)
            with open(DATAAUTHZ_FILENAME, "w") as f1:
                f1.write(dataauthz_render)
            with open(SEARCH_FILENAME, "w") as f2:
                f2.write(search_render)
            with open(MANIFEST_FILENAME, "w") as f3:
                f3.write(manifest_render)
            with tarfile.open(tar_name, "w:gz") as tar_handle:
                tar_handle.add(os.path.abspath(DATAAUTHZ_FILENAME),
                               arcname=DATAAUTHZ_FILENAME)
                tar_handle.add(os.path.abspath(SEARCH_FILENAME),
                               arcname=SEARCH_FILENAME)
                tar_handle.add(os.path.abspath(
                    MANIFEST_FILENAME), arcname=MANIFEST_FILENAME)
            self._upload_to_policy_bucket(tar_name)

            self.log.info(
                f"Uploaded files: {DATAAUTHZ_TEMPLATE_NAME}, {SEARCH_TEMPLATE_NAME}, {MANIFEST_TEMPLATE_NAME} to bundle: {tar_name}")

            os.remove(tar_name)
            os.remove(DATAAUTHZ_FILENAME)
            os.remove(SEARCH_FILENAME)
            os.remove(MANIFEST_FILENAME)


if __name__ == '__main__':
    # Instantiate policy bootstrapping utility
    policy_util = AwsPolicyBootstrap(
        args.region, args.profile, args.policy_bucket, args.partitions, args.debug)

    policy_util.upsert_instance_policy_bundle(DEFAULT_POLICIES_PATH)
    policy_util.upsert_partition_policy_bundles(
        DEFAULT_POLICY_TEMPLATES_PATH)
